package com.skelterlabs.aware

import android.Manifest
import android.app.Activity
import android.content.Context
import androidx.core.app.ActivityCompat
import com.google.gson.JsonParser
import com.skelterlabs.aware.AwarePlugin.CollectorType.BLUETOOTH
import com.skelterlabs.aware.AwarePlugin.CollectorType.CALENDAR
import com.skelterlabs.aware.AwarePlugin.CollectorType.LOCATION
import com.skelterlabs.aware.AwarePlugin.CollectorType.PHOTO_EXIF
import com.skelterlabs.aware.AwarePlugin.CollectorType.SMS
import com.skelterlabs.aware.AwarePlugin.CollectorType.SOUND
import com.skelterlabs.aware.AwarePlugin.CollectorType.WIFI
import com.skelterlabs.aware.AwarePlugin.CollectorType.YOUTUBE
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.FlutterPlugin.FlutterPluginBinding
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry.Registrar
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable
import java.io.IOException
import java.io.InputStreamReader

class AwarePlugin(private val activity: Activity) : MethodCallHandler, FlutterPlugin {
  private var activityContext: Context? = null
  private var methodChannel: MethodChannel? = null
  private var registerDisposable: Disposable? = null
  private var unregisterDisposable: Disposable? = null
  private var profileDisposable: Disposable? = null
  private val aiqAware: AIQAware
    get() {
      return AIQAware.getInstance(activity.application)
    }

  companion object {
    private const val PERMISSION_REQUEST_CODE = 12343
    private const val CLIENT_CONFIG_FILE_NAME = "aware-client-config.json"

    @JvmStatic
    fun registerWith(registrar: Registrar) {
      val instance = AwarePlugin(registrar.activity())
      instance.onAttachedToEngine(registrar.context(), registrar.messenger())
    }
  }

  private enum class CollectorType {
    BLUETOOTH,
    CALENDAR,
    LOCATION,
    PHOTO_EXIF,
    SOUND,
    WIFI,
    YOUTUBE,
    SMS,
    IGNORE,
  }

  private val permissionMap = mapOf(
      BLUETOOTH to listOf(Manifest.permission.BLUETOOTH),
      CALENDAR to listOf(Manifest.permission.READ_CALENDAR),
      LOCATION to listOf(Manifest.permission.ACCESS_FINE_LOCATION),
      PHOTO_EXIF to listOf(Manifest.permission.READ_EXTERNAL_STORAGE),
      SMS to listOf(Manifest.permission.RECEIVE_SMS),
      SOUND to listOf(Manifest.permission.RECORD_AUDIO),
      YOUTUBE to listOf(Manifest.permission.GET_ACCOUNTS),
      WIFI to listOf(
          // See https://developer.android.com/reference/android/net/wifi/WifiManager#getScanResults()
          Manifest.permission.ACCESS_FINE_LOCATION,
          // NOTE: Some devices including Galaxy, Moto and Nokia check this permission and throw
          // exception while wifi scanning.
          Manifest.permission.ACCESS_WIFI_STATE
      )
  )

  private fun convertCollectorNameToType(name: String): CollectorType? {
    return when (name) {
      "bluetooth" -> BLUETOOTH
      "sound" -> SOUND
      "sms" -> SMS
      "calendar" -> CALENDAR
      "location" -> LOCATION
      "photo" -> PHOTO_EXIF
      "wifi" -> WIFI
      "youtube" -> YOUTUBE
      "airPressure",
      "battery",
      "detectedActivity",
      "music",
      "gravity",
      "stepCount" -> null
      else -> null
    }
  }

  override fun onAttachedToEngine(binding: FlutterPluginBinding) {
    onAttachedToEngine(binding.applicationContext, binding.flutterEngine.dartExecutor)
  }

  private fun onAttachedToEngine(context: Context, messenger: BinaryMessenger) {
    aiqAware.init()
    activityContext = context
    methodChannel = MethodChannel(messenger, "aware").apply {
      setMethodCallHandler(this@AwarePlugin)
    }
  }

  override fun onDetachedFromEngine(binding: FlutterPluginBinding) {
    activityContext = null
    methodChannel?.setMethodCallHandler(null)
    methodChannel = null
    registerDisposable?.dispose()
    unregisterDisposable?.dispose()
    profileDisposable?.dispose()
  }

  override fun onMethodCall(call: MethodCall, result: Result) {
    when (call.method) {
      "register" -> register(call, result)
      "unregister" -> unregister(call, result)
      "isRegistered" -> result.success(aiqAware.registered)
      "isServiceEnabled" -> result.success(aiqAware.serviceEnabled)
      "setServiceEnabled" -> setServiceEnabled(call, result)
      "getProfile" -> getProfile(call, result)
      "getClientSecret" -> getClientSecret(call, result)
      "collect" -> collect(call, result)
      else -> result.notImplemented()
    }
  }

  private fun register(call: MethodCall, result: Result) {
    val userHint = call.arguments as? String
    registerDisposable = aiqAware.register(userHint)
        .doOnSuccess {
          registerDisposable?.dispose()
        }
        .observeOn(AndroidSchedulers.mainThread())
        .subscribe(
            { token -> result.success(token) },
            { error -> result.error(call.method, error.localizedMessage, null) })
  }

  private fun unregister(call: MethodCall, result: Result) {
    unregisterDisposable = aiqAware.unregister()
        .doOnComplete { unregisterDisposable?.dispose() }
        .observeOn(AndroidSchedulers.mainThread())
        .subscribe(
            { result.success(true) },
            { error -> result.error(call.method, error.localizedMessage, null) })
  }

  private fun setServiceEnabled(call: MethodCall, result: Result) {
    val enable = call.arguments as Boolean
    aiqAware.serviceEnabled = enable
    result.success(null)
  }

  private fun getProfile(call: MethodCall, result: Result) {
    profileDisposable = aiqAware.getProfile().observeOn(AndroidSchedulers.mainThread())
        .doOnSuccess {
          profileDisposable?.dispose()
        }.subscribe({ profileString ->
          result.success(profileString)
        }, { error ->
          result.error(call.method, error.localizedMessage, null)
        })
  }

  private fun getClientSecret(call: MethodCall, result: Result) {
    val context = activityContext
    if (context == null) {
      result.error(call.method, "There is no context in the aware plugin", null)
      return
    }
    try {
      val inputStream = context.assets.open(CLIENT_CONFIG_FILE_NAME)
      val json = JsonParser().parse(InputStreamReader(inputStream)).asJsonObject
      if (!json.has("api_key")) {
        result.error(call.method, "API key should be set", null)
        return
      }
      result.success(json["api_key"].asString)
    } catch (exception: IOException) {
      result.error(call.method, "There is no $CLIENT_CONFIG_FILE_NAME in assets folder", null)
    }
  }

  private fun collect(call: MethodCall, result: Result) {
    val collectorNames = call.arguments as List<String>

    val collectorTypes = collectorNames.map { name ->
      convertCollectorNameToType(name)
    }

    requestPermissions(collectorTypes.filterNotNull())

    result.success(null)
  }

  private fun requestPermissions(signalTypes: List<CollectorType>) {
    val permissions = signalTypes
        .map { permissionMap.getOrElse(it) { listOf() } }
        .flatten()
        .toTypedArray()

    ActivityCompat.requestPermissions(activity, permissions, PERMISSION_REQUEST_CODE)
  }
}
